from typing import Final
from solders.pubkey import Pubkey

CLOCK: Final[Pubkey]
RECENT_BLOCKHASHES: Final[Pubkey]
RENT: Final[Pubkey]
REWARDS: Final[Pubkey]
STAKE_HISTORY: Final[Pubkey]
EPOCH_SCHEDULE: Final[Pubkey]
INSTRUCTIONS: Final[Pubkey]
SLOT_HASHES: Final[Pubkey]
